︠8b38344f-19fe-4546-8969-ccb34ed2e93a︠
import numpy
import itertools

############################################################################################################################
# For all these procedures we are working with elliptic curve (EC) groups over Z_p where p is a prime number larger than 3.#
# An Elliptic Curve group will be represented by a triple [A,B,p] where                                                    #
#      (1) p is a prime number larger than 3                                                                               #
#      (2) A is the coefficient of x and                                                                                   #
#      (3) B is the constant term in  y^2 = x^3 + A*x + B mod p.                                                           #
#                                                                                                                          #
# A point in the group will be represented as a pair [x,y] where                                                           #
#      (1)  x is the x-coordinate of a solution                                                                            #
#      (2)  y is the y-coordinate of a solution                                                                            #
#      (3)  The identity will be represented by []                                                                         #
############################################################################################################################


def TonSh (a, Prime):
    if kronecker(a, Prime) == -1:
        pass
        print "{0} does not have a square root mod {1}".format(a, Prime)
        return None
    elif Prime % 4 == 3:
        x = power_mod(ZZ(a), ZZ((Prime+1)/4),ZZ(Prime))
        return(x)
    else:
        #################################################################################
        # Determine e so that Prime-1 = (2^e)*q for some odd number q                   # 
        #################################################################################

        e = ZZ(0)
        q = ZZ(Prime - 1)
#        for j in range(1, Prime):
        for j in itertools.count(1):
            if q % 2 == 0:
                e = ZZ(e + 1)
                q = ZZ(q / 2)
            else:
                break
        for i in range(1, 101):
            n = i
            if kronecker(n, Prime) == -1:
                break
        z = power_mod(ZZ(n),ZZ(q),ZZ(Prime))
        y = ZZ(z)
        r = ZZ(e)
        x = power_mod(ZZ(a),ZZ( (q-1)/2),ZZ( Prime))
        b = (a * x ** 2) % Prime
        x = (a * x) % Prime
        #for i in range(1, e + 1):
        for i in itertools.count(1):
            if b == 1:
                break
            else:
                for m in itertools.count(1):
                    t = power_mod(ZZ(b), ZZ(2^m) ,  ZZ(Prime))
                    if t == 1:
                        mm = m
                        break
                t = power_mod(ZZ(y), ZZ(2^(r - mm - 1)),ZZ(Prime))
                y = power_mod(ZZ(t), ZZ(2),ZZ(Prime))
                r = mm
                x = (x * t) % Prime
                b = (b * y) % Prime
        return(x)


︡f10f84b9-6269-4587-bea5-b188da4dd424︡︡{"done":true}
︠a553ebde-f355-41e5-b88c-e13ea00a5c47︠
︡2505fda8-f99b-43d5-ac50-5b6feebffdb5︡︡{"done":true}
︠2c5eaac5-9875-4f22-9ac2-ab341f6ed3e9︠

︡10d88a59-2500-42cd-80a6-2498c66468ff︡︡{"done":true}
︠236c4656-e579-433e-8e97-d3abdb067738i︠
%html
<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between "lowerbound" and "upperbound". The lowerbound and the upperbound are positive random numbers. 
︡eefda683-06b5-4a0c-89e8-4f730397080d︡︡{"done":true,"html":"<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between \"lowerbound\" and \"upperbound\". The lowerbound and the upperbound are positive random numbers."}
︠e6c93c60-6ba8-4bf5-b5db-7557b6152d72︠
import itertools
def ECSearch (lowerbound, upperbound, Group):
    p = Group[2]
    a = Group[0]
    b = Group[1]
    if (4 * a ** 3 + 27 * b ** 2) % p == 0:
        print "This is not an elliptic curve. "
    else:
        for j in itertools.count(lowerbound):
            if j==lowerbound-1 or j>upperbound or j>upperbound:
                return "not found"
            j=j%p
            #print "test "+str(j)
            #print (kronecker(j ** 3 + a * j + b, p) )
            if kronecker(j ** 3 + a * j + b, p) == 1:
                x = (j ** 3 + a * j + b) % p
                var('z')
                #print("stuck here")
                L=TonSh(x,p)
                y = L
                return([j,y])
︡bb665468-f6b2-43a9-82ae-c894619e8b71︡︡{"done":true}
︠8453f9ac-1f94-49c6-a3e1-ae68a4533b47︠

︡677a64c0-ff47-41bd-9071-1532408ad619︡︡{"done":true}
︠de2a39a2-cab7-4b4f-ac69-d3982d5e189f︠

import sympy.ntheory
def FermatGreat (N):
    if sympy.ntheory.isprime(N):
        if N % 4 == 1:
            Z = two_squares(N)
            #print(Z)
            return(Z)
        else:
            print "{0} is prime, but {1} mod 4 = 3".format(N, N)
    else:
        print "{0} is not prime\n".format(N)

︡8d002110-3672-4b86-b047-ea4ec9017ee0︡︡{"done":true}
︠79629385-1c87-4dad-a38f-2d74d025669ei︠

%html
<strong>ECAdd</strong> is a procedure that computes the sum of two points on a given elliptic curve using the elliptic curve operation. 
︡2358f900-5d21-4047-bac4-3c125f016827︡︡{"done":true,"html":"<strong>ECAdd</strong> is a procedure that computes the sum of two points on a given elliptic curve using the elliptic curve operation."}
︠f3b0703d-ea25-4929-a253-57c752ece319︠
︡6b26d192-7130-4be6-a721-4a51e6d140fe︡︡{"done":true}
︠d6ba2d47-db9c-4531-b62c-32bdefe2aa8d︠
def ECAdd (Point1, Point2, Group):
    a = Group[0]
    b = Group[1]
    p = Group[2]
    if(Point1!=[]):
        x1 = Point1[0]
        y1 = Point1[1]
    if(Point2!=[]):
        x2 = Point2[0]
        y2 = Point2[1]
    if (4 * a ** 3 + 27 * b ** 2) % p == 0:
        print "This is not an elliptic curve. "
    elif Point1 != [] and y1 ** 2 % p != (x1 ** 3 + a * x1 + b) % p:
        print "Point 1 is not on the elliptic curve. \n"
    elif Point2 != [] and y2 ** 2 % p != (x2 ** 3 + a * x2 + b) % p:
        print "Point 2 is not on the elliptic curve. \n"
    else:
        if Point1 == []:
            R = Point2
        elif Point2 == []:
            R = Point1
        else:
            if x1 == x2 and 0 == (y1 + y2) % p:
                R = []
            elif x1 == x2 and y1 == y2:
                R = ECDouble(Point1, Group)
            else:
                s = ((y1 - y2) / (x1 - x2)) % p
                x = (s ** 2 - x1 - x2) % p
                y = (s * (x1 - x) - y1) % p
                R = [x,y]
    return(R)
︡3832fa41-fde7-4927-a1e4-4c00b6d65d38︡︡{"done":true}
︠795080ab-d669-44e3-bd82-7d5e0cdd17ae︠
︡6ea77c92-6266-4f0e-8b7f-2f4a64d0bfdb︡︡{"done":true}
︠2c75fed5-4477-4f1b-85e7-a17925d96a6c︠
def ECDouble (Point, Group):
    a = Group[0]
    b = Group[1]
    p = Group[2]
    if Point != []:
        x1 = Point[0]
        y1 = Point[1]
        if (4 * a ** 3 + 27 * b ** 2) % p == 0:
            print "This is not an elliptic curve. "
        elif Point != [] and y1 ** 2 % p != (x1 ** 3 + a * x1 + b) % p:
            print "The point to double is not on the elliptic curve. "
        elif y1 == 0:
            R = []
        else:
            s = mod((3 * x1 ** 2 + a) / y1 / 2, p)
            x = (s ** 2 - 2 * x1) % p
            y = (s * (x1 - x) - y1) % p
            R = [x,y]
    else:
        R = []
    return(R)
︡bdc597e0-0cee-460d-9fbe-4eca9dc0a11b︡︡{"done":true}
︠1e3370a5-ca4b-4c48-9b2f-d5f541a9c1f2︠
︡5b3e8585-fb41-4afa-835d-8527e1db54a9︡︡{"done":true}
︠36c514f0-6055-4771-9023-dd6ec4723281︠

def ECInverse (Point, Group):
    if Point == []:
        return(Point)
    else:
        p = Group[2]
        x = Point[0]
        y = Point[1]
        return([x,(p - y) % p])
︡b32cc71b-e86b-4b1b-9522-2a44049b79a7︡︡{"done":true}
︠248b5919-f1b5-4029-ace3-e1a2a98d2caai︠
%html
<strong>ECTimes</strong> is a procedure that adds a point k-times (scalar) to itself for a given elliptic curve using the elliptic curve operation.
︡18b4f37e-91cf-4637-ba84-cac38e1450c1︡︡{"done":true,"html":"<strong>ECTimes</strong> is a procedure that adds a point k-times (scalar) to itself for a given elliptic curve using the elliptic curve operation."}
︠db90e0c4-4f08-428b-933d-8daa383a99d9︠
def ECTimes (Point, scalar, Group):
    ECIdentity=[]
    if Point == ECIdentity or scalar == 0:
        return(ECIdentity)
    else:
        E=EllipticCurve(GF(Group[2]),[Group[0],Group[1]])
        EPoint = E(Point[0],Point[1])
        #print type(EPoint)
        cgret = scalar*EPoint
        #print cgret[0]
        if(cgret[0]==0 and cgret[1]==1 and cgret[2]==0):
            return ECIdentity
        return([cgret[0],cgret[1]])
 
︡662bbb58-e3ac-4ed5-8ef6-537ede684cf5︡︡{"done":true}
︠17a3578d-39d3-4341-8cfe-ccbbd848b00e︠
G=[3,2,next_prime(11696876878768768767867687687687687678768768766876879878111)]
G ####### Elliptic Curve Group
g=ECSearch(2378648723, 8927492873972973,G)
g ####### base element 
x=34239473274983742
x ####### private key 
b=ECTimes(g,x,G)
︡b9f027bc-7c7b-4f8e-88c9-1b171afbc9bc︡︡{"stdout":"[3, 2, 11696876878768768767867687687687687678768768766876879878131]\n","done":false}︡{"stdout":"[2378648723, 6207257298628741512609567607483777943219211043692587796100]\n","done":false}︡{"stdout":"34239473274983742\n","done":false}︡{"done":true}
︠8cc6cad6-173b-4def-b88d-42c064bb67d7︠
︡646962f6-9d03-47a8-9bab-f272b75a09a7︡︡{"done":true}
︠d6ddec34-951c-4618-9eb5-deb24ee08189︠

def listproduct (primelist):
    x = primelist
    k = len(primelist)
    print k
    result = 1
    for i in range(0, k):
        result = result * x[i][0] ^ x[i][1] #op(1, op(i, x)) ** op(2, op(i, x))
    return(result)

def listptorder (pt, Group, factoredGpOrder):
    ECIdentity = []
    x = factoredGpOrder
    k = len(factoredGpOrder)
    orderlist = set([])
    result = listproduct(factoredGpOrder)
    for i in range(0, k ):
        p =  x[i][0] #op(1, op(i, x))
        a =  x[i][1] #op(2, op(i, x))
        ord = 0
        for j in range(0, a ):
            result = result / p
            test = ECTimes(pt, result, Group)
            if test != ECIdentity:
                result = result * p
                ord = ord + 1
        if 0 < ord:
            orderlist = orderlist | set([[p,ord]])
    return(convert(orderlist, list))


def primeHPSonEC (y,g,Group,p):
    newg = []#ECIdentity
    for j in range(0, p ):
        if y == newg:
            break
        newg = ECAdd(newg, g, Group)
    return(j)


def powerHPSonEC (y,g,og,p,a,Group):
    gp = ECTimes(g, og / p, Group)
    newog = og
    newy = y
    c = 0
    partx = 0
    for i in range(0, a ):
        newy = ECAdd(y, ECInverse(ECTimes(g, partx, Group), Group), Group)
        newog = newog / p
        ypower = ECTimes(newy, newog, Group)
        c = primeHPSonEC(ypower, gp, Group, p)
        partx = partx + c * p ** i
    return(partx)

def HPSonEC (y,g,Group,factoredGpOrder):
    X = 0
    Ord = 1
    K = listptorder(g, Group, factoredGpOrder)
    k = len(K)
    ordg = listproduct(K)
    for j in range(1, k + 1):
        p = K[j][0]#op(1, op(j, K))
        powerp = K[j][1]#op(2, op(j, K))
        partx = powerHPSonEC(y, g, ordg, p, powerp, Group)
        if j == 1:
            X = partx
            Ord = p ** powerp
        else:
            X = crt([X,partx], [Ord,p ** powerp])
            Ord = Ord * p ** powerp
    return(X)
︡1a29f1cb-3430-4691-8f5b-51f694b9abcc︡︡{"done":true}
︠4f3ec723-2096-4afd-bded-0343200042b0︠
import math

def ASCIIPad(Mess,Mod):
    K = []
    for letter in Mess:
        K.append(ord(letter))
    L = Mod.ndigits()
    l = len(K)
    y = ZZ(math.floor(L/3))
    count = 0
    padded = []
    buffer = ""
    for numChar in K:
        numChar+=100
        buffer+=str(numChar)
        count+=1
        if count==y:
            padded.append(ZZ(buffer))
            buffer = ""
            count = 0
    if len(buffer)>0:
        padded.append(ZZ(buffer))
    return padded

                
def ASCIIDepad(Number):
    N = "";
    #Number = ZZ(Number[0])
    n = Number.ndigits() % 3;
    if (n > 0):
        print("This is not a padded ASCII string\n");
    else:
        L = [((Number - (Number % (1000^i)))/1000^i)%1000 - 100 for i in range(Number.ndigits()/3)];
        for i in range(Number.ndigits()/3):
            #print L[i]
            N = chr(L[i]) + N;
            
    return(N);

def ECEmbed (Message, gp, tolerance):
    p = ZZ(math.floor(gp[2] / (tolerance + 1)))
    M = ASCIIPad(Message, p)
    packets = len(M)
    #print packets
    pointlist = [0]*packets
    for j in range(0, packets ):
        N = M[j]
#        N:=(op(0,op(1,M))[j]);
        pointlist[j] = ECSearch(tolerance * N, tolerance * (N + 1) - 1, gp)
        #print pointlist
    return(pointlist)

def ECUnembed (pointlist, tolerance):
    #print pointlist
    k = len(pointlist)
    paddedasciilist=[0]*k
    for j in range(0, k ):
        #print type(pointlist[j][0]/tolerance)
        #print pointlist
        #print paddedasciilist        
        #print "test this "+str((pointlist[j][0]/tolerance))
        #.floor() returns correct while math.floor() does not work
        #print (pointlist[j][0])
        pointlist[j][0]=ZZ(pointlist[j][0])
        toType = ZZ(QQ((pointlist[j][0])/tolerance).floor())
#        typed = type(toType)
#        print typed
        paddedasciilist[j] = ((pointlist[j][0])/tolerance).floor()
    returnStr = ""
    for paddedItem in paddedasciilist:
        #print paddedItem
        buffer = ASCIIDepad(paddedItem)
        returnStr+= buffer
        #print buffer
    return returnStr
︡0e1e7b87-cf27-47fd-a854-673d6f730623︡︡{"done":true}
︠d346240a-1f2b-4781-ad9e-9a9a322ad4aci︠
%html
<strong>Example.</strong> Create an elliptic curve public and private key. 
   


︡a54d4bdb-d93b-46e7-a312-0664cab0027a︡︡{"done":true,"html":"<strong>Example.</strong> Create an elliptic curve public and private key."}
︠acc94600-2445-4afd-9e19-ee572963e577︠
G=[3,2,next_prime(1169687687876876876786768768768768767876876876687687987811546546545646541)]
G ####### Elliptic Curve Group
g=ECSearch(2378648723, 8927492873972973,G)
g ####### base element 
x=34239473274983742
x ####### private key 
b=ECTimes(g,x,G)
b
︡05a8a65a-9a5b-459d-88d2-5642e2f0426d︡︡{"stdout":"[3, 2, 1169687687876876876786768768768768767876876876687687987811546546545646659]\n","done":false}︡{"stdout":"[2378648723, 251129725725808360916423247622071249545935346397519406481708557006516860]\n","done":false}︡{"stdout":"34239473274983742\n","done":false}︡{"stdout":"[102185937035780644447440724514674909408114529521257498878576073830758885, 29294782479764415939896071953678258471598161093478693795502004627180723]\n","done":false}︡{"done":true}
︠18a1f4ba-3253-4d6c-9194-4e2d582528b5i︠
%html
<strong>Example.</strong> Encrypt the plaintext "Cryptology" using the the elliptic curve and the key created in the previous example. The tolerance parameter is T=40.
︡0ae817fa-799d-4813-a0bc-d79b594be39a︡︡{"done":true,"html":"<strong>Example.</strong> Encrypt the plaintext \"Cryptology\" using the the elliptic curve and the key created in the previous example. The tolerance parameter is T=40."}
︠834fd10b-2992-4695-a42a-8ec93bda5b0f︠

m=ECEmbed("Cryptology",G, 40)
m ###### embedded plaintext into an elliptic curve group
r=34239473274983742
r ##### random number
h=ECTimes(g,r,G)
h
s=ECTimes(b,r,G)
s
c=ECAdd(m[0],s,G)
c
################################## The CIPHERTEXT is the pair of group elements h and c #####################################
︡8327485a-287d-4dc1-b128-045895eb684a︡︡{"stdout":"[[6688568848488648448328448128842, 30713016421388513534069392095726691904699597855212477624228805094571897]]\n","done":false}︡{"stdout":"34239473274983742\n","done":false}︡{"stdout":"[102185937035780644447440724514674909408114529521257498878576073830758885, 29294782479764415939896071953678258471598161093478693795502004627180723]\n","done":false}︡{"stdout":"[531594751936260343069237726792546775307287241364648309851908704952472378, 62836598762052693815835015170177315163174591233643267255681661239022069]\n","done":false}︡{"stdout":"[60725488831383488099275639299974627203302101759526710107542683749459307, 786886098612758367049369768966181063935118105698230786854120868069958151]\n","done":false}︡{"done":true}
︠7dfaefd0-dc95-44d3-9384-917264a09bc3︠

︡e28c4e49-b2fe-45ea-ad24-0bc33f66f2dd︡︡{"done":true}
︠2c128a5e-ed7b-4f5d-9559-4fff46ca7509︠
%html
<strong>Example.</strong> Decrypt the ciphertext from the previous example. IN CLASS!
︡97d77467-d0ee-431f-a6a1-279c5aa1b802︡︡{"done":true,"html":"<strong>Example.</strong> Decrypt the ciphertext from the previous example. IN CLASS!"}
︠265d441b-dee9-4d80-b062-3b3852bb48b0︠

︡032173d0-df9d-478d-a8c0-b80d4926e736︡︡{"done":true}
︠ec682e34-8217-4632-9720-9d69c477b6fbi︠
%html
<strong>Example.</strong> Encryption of a plaintext represented by a list of points in the elliptic curve group G.
︡fa07d336-70a8-487a-abb3-c4c1da618208︡︡{"done":true,"html":"<strong>Example.</strong> Encryption of a plaintext represented by a list of points in the elliptic curve group G."}
︠b426f33d-6780-403f-8a73-b833181b2a89︠
#########################################################################################################################
# Note: When running this example make sure you run them in order so that other cell's definitions are not used by sage # 
#########################################################################################################################

m=ECEmbed("This is a longer example. It uses multiple points. The points are listed right below and then the random number chosen and the last output is how many points are needed.",G, 40)
m ###### embedded plaintext into an elliptic curve group
r=3423947327498374254646554654645
r ##### random number
numPoints = len(m)
numPoints
h=[0]*numPoints
s=[0]*numPoints
c=[0]*numPoints
for counter in xrange(numPoints):
    h=ECTimes(g,r,G)
    s=ECTimes(b,r,G)
    c[counter]=ECAdd(m[counter],s,G)

︡a244914c-80ef-4ffa-a7c1-f9ec0d7dc1e7︡︡{"stdout":"[[7368168208605288208605287885288328448408128048565288048807888368488320, 121486482953176117528000809881749947571671902330970782925843120215624872], [8045845286928645288688608048605288368688328648208488328045288488448200, 903695374841292322370271174415204218208659055675797121100003842392500926], [8408648605845287368168045288488448208408648605287888568045288328208600, 452563156195183222674654981175834124068138520541249589841058897891287244], [8648048005288568208128168645287928048328448765287888408005288648168040, 354133454203841855971589741317387587553570850607256693528118869301689901], [8405288648168045288567888408008448365288408688367928048565287968168440, 566116327420959706426335734957260808053541488054698602617229920903468133], [8608048405287888408005288648168045288327888608645288448688648488688641, 190989246142380934866236751363212697612118794496744686603811386587621417], [5288208605288168448765288367888408845288488448208408648605287888568041, 425171620928964741474599056379821932316091568940530183178971060551595853], [5288408048048008048005840, 1036270550319895031743093327765070280126725147090778246179870977799274707]]\n","done":false}︡{"stdout":"3423947327498374254646554654645\n","done":false}︡{"stdout":"8\n","done":false}︡{"done":true}
︠9a7286b8-516d-4e5c-a14f-4d08f7e70446︠
h
s
c
︡b35c64ab-4f90-48be-a5b4-dd2cade69bf9︡︡{"stdout":"[502800681108808127999931477346298640902458497799276570157831511464842627, 249433093957737059630763632764052800806824499724831296249200563703070996]\n","done":false}︡{"stdout":"[758836381415980725873880391356898132106948925841767677664020161182662238, 172885080124927030627775664387974993851680223072695154225919344264781309]\n","done":false}︡{"stdout":"[[993011997632563665090520795819224461956877820256329108993458876624168267, 486967488472059785832649245344787601820108477716142036228174305706008047], [1025130127158111391286114497628784310721816018761263921410307202662962155, 1046214728711544478045606234428908877874990102826941182072699504231466650], [702163857070541572271486478402213299769149650327011285713214313411220209, 964390034282681070472559926854266573294570007713258722688970731401446070], [545974440942587589793355643326618406027250242128080874758407596428510512, 761850948586179337743907452723161818650832861623511156392451479702387930], [1007560043455625957074409098454581056661266112007216804759867237444073697, 568301847329160573315400903846138202220379078132258736522271094761757556], [197721450177956824925214814076136877509231101020546157774722389794437045, 364710051200340597882807263315974397500183542853128578183482144214813413], [1040787630887710087842592894781064728294150449155777473810980176110345869, 209262206482771129037093161089704974346792207851381833556910011562779092], [817585719106595758704863067855703256513213471986817100166846551663829691, 537939259557282367210288215732191700156429531225372888960276076554578723]]\n","done":false}︡{"done":true}
︠7a3cd21c-aa84-485a-bb72-7cde67b01067︠
z=ECTimes(h,x,G)
z
d = [0]*len(c)
for counter in xrange(len(c)):
    d[counter] = ECAdd(c[counter],ECInverse(z,G),G)
d
ECUnembed(d,40)
︡9b831534-efd0-4c22-86e5-6236798bb11e︡︡{"stdout":"[758836381415980725873880391356898132106948925841767677664020161182662238, 172885080124927030627775664387974993851680223072695154225919344264781309]\n","done":false}︡{"stdout":"[[7368168208605288208605287885288328448408128048565288048807888368488320, 121486482953176117528000809881749947571671902330970782925843120215624872], [8045845286928645288688608048605288368688328648208488328045288488448200, 903695374841292322370271174415204218208659055675797121100003842392500926], [8408648605845287368168045288488448208408648605287888568045288328208600, 452563156195183222674654981175834124068138520541249589841058897891287244], [8648048005288568208128168645287928048328448765287888408005288648168040, 354133454203841855971589741317387587553570850607256693528118869301689901], [8405288648168045288567888408008448365288408688367928048565287968168440, 566116327420959706426335734957260808053541488054698602617229920903468133], [8608048405287888408005288648168045288327888608645288448688648488688641, 190989246142380934866236751363212697612118794496744686603811386587621417], [5288208605288168448765288367888408845288488448208408648605287888568041, 425171620928964741474599056379821932316091568940530183178971060551595853], [5288408048048008048005840, 1036270550319895031743093327765070280126725147090778246179870977799274707]]\n","done":false}︡{"stdout":"'This is a longer example. It uses multiple points. The points are listed right below and then the random number chosen and the last output is how many points are needed.'\n","done":false}︡{"done":true}
︠7451b2dd-93e8-4bf5-9b2e-1b30edc8139e︠

︡15092d6b-eeeb-462e-b885-8aafe1e22db4︡︡{"done":true}
︠98f3b14d-4cd9-4a23-825e-103c8dc030b9︠


︡d45ab8f2-ae84-4891-b088-b6db92f3b9b5︡︡{"done":true}
︠aa2c8dee-ff2b-41f5-82b1-1a754d5725a6︠









︡b69fec0b-02d9-4249-89d4-8d6269e63442︡
︠e5ac8094-764e-4dfc-94c8-c504c368f9e9︠
import numpy
import itertools

############################################################################################################################
# For all these procedures we are working with elliptic curve (EC) groups over Z_p where p is a prime number larger than 3.#
# An Elliptic Curve group will be represented by a triple [A,B,p] where                                                    #
#      (1) p is a prime number larger than 3                                                                               #
#      (2) A is the coefficient of x and                                                                                   #
#      (3) B is the constant term in  y^2 = x^3 + A*x + B mod p.                                                           #
#                                                                                                                          #
# A point in the group will be represented as a pair [x,y] where                                                           #
#      (1)  x is the x-coordinate of a solution                                                                            #
#      (2)  y is the y-coordinate of a solution                                                                            #
#      (3)  The identity will be represented by []                                                                         #
############################################################################################################################


def TonSh (a, Prime):
    if kronecker(a, Prime) == -1:
        pass
        print "{0} does not have a square root mod {1}".format(a, Prime)
        return None
    elif Prime % 4 == 3:
        x = power_mod(ZZ(a), ZZ((Prime+1)/4),ZZ(Prime))
        return(x)
    else:
        #################################################################################
        # Determine e so that Prime-1 = (2^e)*q for some odd number q                   # 
        #################################################################################

        e = ZZ(0)
        q = ZZ(Prime - 1)
#        for j in range(1, Prime):
        for j in itertools.count(1):
            if q % 2 == 0:
                e = ZZ(e + 1)
                q = ZZ(q / 2)
            else:
                break
        for i in range(1, 101):
            n = i
            if kronecker(n, Prime) == -1:
                break
        z = power_mod(ZZ(n),ZZ(q),ZZ(Prime))
        y = ZZ(z)
        r = ZZ(e)
        x = power_mod(ZZ(a),ZZ( (q-1)/2),ZZ( Prime))
        b = (a * x ** 2) % Prime
        x = (a * x) % Prime
        #for i in range(1, e + 1):
        for i in itertools.count(1):
            if b == 1:
                break
            else:
                for m in itertools.count(1):
                    t = power_mod(ZZ(b), ZZ(2^m) ,  ZZ(Prime))
                    if t == 1:
                        mm = m
                        break
                t = power_mod(ZZ(y), ZZ(2^(r - mm - 1)),ZZ(Prime))
                y = power_mod(ZZ(t), ZZ(2),ZZ(Prime))
                r = mm
                x = (x * t) % Prime
                b = (b * y) % Prime
        return(x)


︡d9d31de0-e2bc-437d-b411-fd1c3141d682︡︡{"done":true}
︠e3235d6b-2558-49c2-9a99-5630eda30b39︠
︡da428c89-418c-4331-8208-4afd3643f3a1︡︡{"done":true}
︠31522e96-e812-47e6-badc-4802bde77d4b︠

︡a74a3478-2613-4d4c-9e38-4a00a702797c︡︡{"done":true}
︠f9125ca3-40a1-4b1b-a1bb-3df87c05fd19i︠
%html
<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between "lowerbound" and "upperbound". The lowerbound and the upperbound are positive random numbers. 
︡f8300351-d85f-46d0-b3a8-d6bec8245568︡︡{"done":true,"html":"<strong>ECSearch</strong> is a procedure that searches for a point on a given elliptic curve with first coordinate between \"lowerbound\" and \"upperbound\". The lowerbound and the upperbound are positive random numbers."}
︠dffbc1c0-4e81-44a9-a004-2bd3e1e443c7︠
import itertools
def ECSearch (lowerbound, upperbound, Group):
    p = Group[2]
    a = Group[0]
    b = Group[1]
    if (4 * a ** 3 + 27 * b ** 2) % p == 0:
        print "This is not an elliptic curve. "
    else:
        for j in itertools.count(lowerbound):
            if j==lowerbound-1 or j>upperbound or j>upperbound:
                return "not found"
            j=j%p
            #print "test "+str(j)
            #print (kronecker(j ** 3 + a * j + b, p) )
            if kronecker(j ** 3 + a * j + b, p) == 1:
                x = (j ** 3 + a * j + b) % p
                var('z')
                #print("stuck here")
                L=TonSh(x,p)
                y = L
                return([j,y])
︡96715a11-4b97-4cde-811d-06bb19d66f1d︡︡{"done":true}
︠905bc73d-1dee-4000-b542-d3326278d236︠

︡a01a413b-7c94-42c5-9f7a-57e92933961c︡︡{"done":true}
︠5085d8e3-748a-4300-b131-c9bc99023857︠

import sympy.ntheory
def FermatGreat (N):
    if sympy.ntheory.isprime(N):
        if N % 4 == 1:
            Z = two_squares(N)
            #print(Z)
            return(Z)
        else:
            print "{0} is prime, but {1} mod 4 = 3".format(N, N)
    else:
        print "{0} is not prime\n".format(N)

︡3a38e482-ae68-45cb-879e-70ad48a8b435︡︡{"done":true}
︠27226fb1-3a77-4c55-8c8f-e28f69bc1fdci︠

%html
<strong>ECAdd</strong> is a procedure that computes the sum of two points on a given elliptic curve using the elliptic curve operation. 
︡fc8802f4-b3a6-48ef-8f15-34c79e9bb306︡︡{"done":true,"html":"<strong>ECAdd</strong> is a procedure that computes the sum of two points on a given elliptic curve using the elliptic curve operation."}
︠653b5193-95f3-43b9-a8ec-6cff4ac0c838︠
︡67e59c0c-e75a-443f-b119-ad201f8ba672︡︡{"done":true}
︠c336b9d9-6dfa-40fc-86c4-25e86be82df5︠
def ECAdd (Point1, Point2, Group):
    a = Group[0]
    b = Group[1]
    p = Group[2]
    if(Point1!=[]):
        x1 = Point1[0]
        y1 = Point1[1]
    if(Point2!=[]):
        x2 = Point2[0]
        y2 = Point2[1]
    if (4 * a ** 3 + 27 * b ** 2) % p == 0:
        print "This is not an elliptic curve. "
    elif Point1 != [] and y1 ** 2 % p != (x1 ** 3 + a * x1 + b) % p:
        print "Point 1 is not on the elliptic curve. \n"
    elif Point2 != [] and y2 ** 2 % p != (x2 ** 3 + a * x2 + b) % p:
        print "Point 2 is not on the elliptic curve. \n"
    else:
        if Point1 == []:
            R = Point2
        elif Point2 == []:
            R = Point1
        else:
            if x1 == x2 and 0 == (y1 + y2) % p:
                R = []
            elif x1 == x2 and y1 == y2:
                R = ECDouble(Point1, Group)
            else:
                s = ((y1 - y2) / (x1 - x2)) % p
                x = (s ** 2 - x1 - x2) % p
                y = (s * (x1 - x) - y1) % p
                R = [x,y]
    return(R)
︡00b0dcbb-e705-469e-90f1-dd2c0f43617b︡︡{"done":true}
︠cf406f3d-bd4b-489f-b5aa-6b12a4a5663d︠
︡eb1bbb48-0397-43bf-8519-8fe289cc404e︡︡{"done":true}
︠34583d61-7f5f-45eb-b465-761018fb710b︠
def ECDouble (Point, Group):
    a = Group[0]
    b = Group[1]
    p = Group[2]
    if Point != []:
        x1 = Point[0]
        y1 = Point[1]
        if (4 * a ** 3 + 27 * b ** 2) % p == 0:
            print "This is not an elliptic curve. "
        elif Point != [] and y1 ** 2 % p != (x1 ** 3 + a * x1 + b) % p:
            print "The point to double is not on the elliptic curve. "
        elif y1 == 0:
            R = []
        else:
            s = mod((3 * x1 ** 2 + a) / y1 / 2, p)
            x = (s ** 2 - 2 * x1) % p
            y = (s * (x1 - x) - y1) % p
            R = [x,y]
    else:
        R = []
    return(R)
︡251cb94b-8541-4c25-a3e9-d670573a31d9︡︡{"done":true}
︠6a0b4957-ed67-4659-9213-f4828b364201︠
︡5cc15fdd-8ae7-4fde-9f8f-5ddb23e75cfb︡︡{"done":true}
︠ea687121-cd05-4ba0-a1bc-8b27d97bad89︠

def ECInverse (Point, Group):
    if Point == []:
        return(Point)
    else:
        p = Group[2]
        x = Point[0]
        y = Point[1]
        return([x,(p - y) % p])
︡ad8da853-a1e8-4cd6-953e-4c941f9752d5︡︡{"done":true}
︠d9bdc58b-3dd2-4c16-b166-7c9ea512074ei︠
%html
<strong>ECTimes</strong> is a procedure that adds a point k-times (scalar) to itself for a given elliptic curve using the elliptic curve operation.
︡2c71ac89-ad3a-4d9b-8176-9063ba9a47ac︡︡{"done":true,"html":"<strong>ECTimes</strong> is a procedure that adds a point k-times (scalar) to itself for a given elliptic curve using the elliptic curve operation."}
︠25bd87ce-dbcf-431d-a0f4-be6c96b734ab︠
def ECTimes (Point, scalar, Group):
    ECIdentity=[]
    if Point == ECIdentity or scalar == 0:
        return(ECIdentity)
    else:
        E=EllipticCurve(GF(Group[2]),[Group[0],Group[1]])
        EPoint = E(Point[0],Point[1])
        #print type(EPoint)
        cgret = scalar*EPoint
        #print cgret[0]
        if(cgret[0]==0 and cgret[1]==1 and cgret[2]==0):
            return ECIdentity
        return([cgret[0],cgret[1]])
 
︡ad569711-f4b0-4282-a951-3e6949fc2b63︡︡{"done":true}
︠f2baef69-cf30-4049-a7dd-21ae0f5bd293︠
G=[3,2,next_prime(11696876878768768767867687687687687678768768766876879878111)]
G ####### Elliptic Curve Group
g=ECSearch(2378648723, 8927492873972973,G)
g ####### base element 
x=34239473274983742
x ####### private key 
b=ECTimes(g,x,G)
︡761bcf10-2258-41ef-a6ff-63ffb9c3531b︡︡{"stdout":"[3, 2, 11696876878768768767867687687687687678768768766876879878131]\n","done":false}︡{"stdout":"[2378648723, 6207257298628741512609567607483777943219211043692587796100]\n","done":false}︡{"stdout":"34239473274983742\n","done":false}︡{"done":true}
︠cd2ae02b-c7d1-4e52-830f-ead16b12d3fa︠
︡615cdd37-66a9-47ba-abad-47830646b7f9︡︡{"done":true}
︠624efc8a-c5c5-4f28-97dc-2ed40d3c099f︠

def listproduct (primelist):
    x = primelist
    k = len(primelist)
    print k
    result = 1
    for i in range(0, k):
        result = result * x[i][0] ^ x[i][1] #op(1, op(i, x)) ** op(2, op(i, x))
    return(result)

def listptorder (pt, Group, factoredGpOrder):
    ECIdentity = []
    x = factoredGpOrder
    k = len(factoredGpOrder)
    orderlist = set([])
    result = listproduct(factoredGpOrder)
    for i in range(0, k ):
        p =  x[i][0] #op(1, op(i, x))
        a =  x[i][1] #op(2, op(i, x))
        ord = 0
        for j in range(0, a ):
            result = result / p
            test = ECTimes(pt, result, Group)
            if test != ECIdentity:
                result = result * p
                ord = ord + 1
        if 0 < ord:
            orderlist = orderlist | set([[p,ord]])
    return(convert(orderlist, list))


def primeHPSonEC (y,g,Group,p):
    newg = []#ECIdentity
    for j in range(0, p ):
        if y == newg:
            break
        newg = ECAdd(newg, g, Group)
    return(j)


def powerHPSonEC (y,g,og,p,a,Group):
    gp = ECTimes(g, og / p, Group)
    newog = og
    newy = y
    c = 0
    partx = 0
    for i in range(0, a ):
        newy = ECAdd(y, ECInverse(ECTimes(g, partx, Group), Group), Group)
        newog = newog / p
        ypower = ECTimes(newy, newog, Group)
        c = primeHPSonEC(ypower, gp, Group, p)
        partx = partx + c * p ** i
    return(partx)

def HPSonEC (y,g,Group,factoredGpOrder):
    X = 0
    Ord = 1
    K = listptorder(g, Group, factoredGpOrder)
    k = len(K)
    ordg = listproduct(K)
    for j in range(1, k + 1):
        p = K[j][0]#op(1, op(j, K))
        powerp = K[j][1]#op(2, op(j, K))
        partx = powerHPSonEC(y, g, ordg, p, powerp, Group)
        if j == 1:
            X = partx
            Ord = p ** powerp
        else:
            X = crt([X,partx], [Ord,p ** powerp])
            Ord = Ord * p ** powerp
    return(X)
︡61cddba6-548d-4103-ab01-2f8b50b4f165︡︡{"done":true}
︠8d1b034b-c964-41d3-ad66-6f05f2003054︠
import math

def ASCIIPad(Mess,Mod):
    K = []
    for letter in Mess:
        K.append(ord(letter))
    L = Mod.ndigits()
    l = len(K)
    y = ZZ(math.floor(L/3))
    count = 0
    padded = []
    buffer = ""
    for numChar in K:
        numChar+=100
        buffer+=str(numChar)
        count+=1
        if count==y:
            padded.append(ZZ(buffer))
            buffer = ""
            count = 0
    if len(buffer)>0:
        padded.append(ZZ(buffer))
    return padded

                
def ASCIIDepad(Number):
    N = "";
    #Number = ZZ(Number[0])
    n = Number.ndigits() % 3;
    if (n > 0):
        print("This is not a padded ASCII string\n");
    else:
        L = [((Number - (Number % (1000^i)))/1000^i)%1000 - 100 for i in range(Number.ndigits()/3)];
        for i in range(Number.ndigits()/3):
            #print L[i]
            N = chr(L[i]) + N;
            
    return(N);

def ECEmbed (Message, gp, tolerance):
    p = ZZ(math.floor(gp[2] / (tolerance + 1)))
    M = ASCIIPad(Message, p)
    packets = len(M)
    #print packets
    pointlist = [0]*packets
    for j in range(0, packets ):
        N = M[j]
#        N:=(op(0,op(1,M))[j]);
        pointlist[j] = ECSearch(tolerance * N, tolerance * (N + 1) - 1, gp)
        #print pointlist
    return(pointlist)

def ECUnembed (pointlist, tolerance):
    #print pointlist
    k = len(pointlist)
    paddedasciilist=[0]*k
    for j in range(0, k ):
        #print type(pointlist[j][0]/tolerance)
        #print pointlist
        #print paddedasciilist        
        #print "test this "+str((pointlist[j][0]/tolerance))
        #.floor() returns correct while math.floor() does not work
        #print (pointlist[j][0])
        pointlist[j][0]=ZZ(pointlist[j][0])
        toType = ZZ(QQ((pointlist[j][0])/tolerance).floor())
#        typed = type(toType)
#        print typed
        paddedasciilist[j] = ((pointlist[j][0])/tolerance).floor()
    returnStr = ""
    for paddedItem in paddedasciilist:
        #print paddedItem
        buffer = ASCIIDepad(paddedItem)
        returnStr+= buffer
        #print buffer
    return returnStr
︡ac1e0ef1-b78c-4d44-9433-c74b17484f27︡︡{"done":true}
︠7f76b21d-046f-423b-96a5-14e521cf13a1i︠
%html
<strong>Example.</strong> Create an elliptic curve public and private key. 
   


︡6a247f49-959c-4d03-890a-f29fe2afd8a1︡︡{"done":true,"html":"<strong>Example.</strong> Create an elliptic curve public and private key."}
︠e0938244-1a02-474a-bfcc-f13a0f158c6a︠
G=[3,2,next_prime(1169687687876876876786768768768768767876876876687687987811546546545646541)]
G ####### Elliptic Curve Group
g=ECSearch(2378648723, 8927492873972973,G)
g ####### base element 
x=34239473274983742
x ####### private key 
b=ECTimes(g,x,G)
b
︡f60cd521-b6b5-4be6-951b-0f2b6dcafb91︡︡{"stdout":"[3, 2, 1169687687876876876786768768768768767876876876687687987811546546545646659]\n","done":false}︡{"stdout":"[2378648723, 251129725725808360916423247622071249545935346397519406481708557006516860]\n","done":false}︡{"stdout":"34239473274983742\n","done":false}︡{"stdout":"[102185937035780644447440724514674909408114529521257498878576073830758885, 29294782479764415939896071953678258471598161093478693795502004627180723]\n","done":false}︡{"done":true}
︠189e8442-bfde-4226-a3ab-7754528a5295i︠
%html
<strong>Example.</strong> Encrypt the plaintext "Cryptology" using the the elliptic curve and the key created in the previous example. The tolerance parameter is T=40.
︡94ac3329-6eb4-42a2-b872-788e4d5546e1︡︡{"done":true,"html":"<strong>Example.</strong> Encrypt the plaintext \"Cryptology\" using the the elliptic curve and the key created in the previous example. The tolerance parameter is T=40."}
︠02ec455f-ff94-4bb8-a2db-e46a1cf4c3e4︠

m=ECEmbed("Cryptology",G, 40)
m ###### embedded plaintext into an elliptic curve group
r=34239473274983742
r ##### random number
h=ECTimes(g,r,G)
h
s=ECTimes(b,r,G)
s
c=ECAdd(m[0],s,G)
c
################################## The CIPHERTEXT is the pair of group elements h and c #####################################
︡67cea1da-b7b7-4b9f-a999-f4f3d4397dce︡︡{"stdout":"[[6688568848488648448328448128842, 30713016421388513534069392095726691904699597855212477624228805094571897]]\n","done":false}︡{"stdout":"34239473274983742\n","done":false}︡{"stdout":"[102185937035780644447440724514674909408114529521257498878576073830758885, 29294782479764415939896071953678258471598161093478693795502004627180723]\n","done":false}︡{"stdout":"[531594751936260343069237726792546775307287241364648309851908704952472378, 62836598762052693815835015170177315163174591233643267255681661239022069]\n","done":false}︡{"stdout":"[60725488831383488099275639299974627203302101759526710107542683749459307, 786886098612758367049369768966181063935118105698230786854120868069958151]\n","done":false}︡{"done":true}
︠b6558e06-3c1d-43fc-8e6b-858dc61a9184︠

︡a7dad944-6a8b-404b-8e6c-3d08ba852580︡︡{"done":true}
︠68af939e-0273-4af1-a516-95b613403a54︠
%html
<strong>Example.</strong> Decrypt the ciphertext from the previous example. IN CLASS!
︡300bf152-8533-497a-8728-0a8260285ec4︡︡{"done":true,"html":"<strong>Example.</strong> Decrypt the ciphertext from the previous example. IN CLASS!"}
︠3ec7e0cd-fbe5-4f75-8b39-419152c3352c︠

︡c1e9d97e-6330-4aa2-aaef-5136064b49b1︡︡{"done":true}
︠58b1930b-56fa-40bf-b138-7bd1c5622c4fi︠
%html
<strong>Example.</strong> Encryption of a plaintext represented by a list of points in the elliptic curve group G.
︡2c159abf-a53c-4d76-b603-9d80e8329bf5︡︡{"done":true,"html":"<strong>Example.</strong> Encryption of a plaintext represented by a list of points in the elliptic curve group G."}
︠fe77f364-6f80-46d5-bb79-cdadeb485118︠
#########################################################################################################################
# Note: When running this example make sure you run them in order so that other cell's definitions are not used by sage # 
#########################################################################################################################

m=ECEmbed("This is a longer example. It uses multiple points. The points are listed right below and then the random number chosen and the last output is how many points are needed.",G, 40)
m ###### embedded plaintext into an elliptic curve group
r=3423947327498374254646554654645
r ##### random number
numPoints = len(m)
numPoints
h=[0]*numPoints
s=[0]*numPoints
c=[0]*numPoints
for counter in xrange(numPoints):
    h=ECTimes(g,r,G)
    s=ECTimes(b,r,G)
    c[counter]=ECAdd(m[counter],s,G)

︡d835c699-a8ba-4feb-b7e1-449fe9d42403︡︡{"stdout":"[[7368168208605288208605287885288328448408128048565288048807888368488320, 121486482953176117528000809881749947571671902330970782925843120215624872], [8045845286928645288688608048605288368688328648208488328045288488448200, 903695374841292322370271174415204218208659055675797121100003842392500926], [8408648605845287368168045288488448208408648605287888568045288328208600, 452563156195183222674654981175834124068138520541249589841058897891287244], [8648048005288568208128168645287928048328448765287888408005288648168040, 354133454203841855971589741317387587553570850607256693528118869301689901], [8405288648168045288567888408008448365288408688367928048565287968168440, 566116327420959706426335734957260808053541488054698602617229920903468133], [8608048405287888408005288648168045288327888608645288448688648488688641, 190989246142380934866236751363212697612118794496744686603811386587621417], [5288208605288168448765288367888408845288488448208408648605287888568041, 425171620928964741474599056379821932316091568940530183178971060551595853], [5288408048048008048005840, 1036270550319895031743093327765070280126725147090778246179870977799274707]]\n","done":false}︡{"stdout":"3423947327498374254646554654645\n","done":false}︡{"stdout":"8\n","done":false}︡{"done":true}
︠c5e95904-9513-4cdf-bb41-a9237be49404︠
h
s
c
︡de97c4b6-7fc6-4f2e-b293-b2fa19b3f908︡︡{"stdout":"[502800681108808127999931477346298640902458497799276570157831511464842627, 249433093957737059630763632764052800806824499724831296249200563703070996]\n","done":false}︡{"stdout":"[758836381415980725873880391356898132106948925841767677664020161182662238, 172885080124927030627775664387974993851680223072695154225919344264781309]\n","done":false}︡{"stdout":"[[993011997632563665090520795819224461956877820256329108993458876624168267, 486967488472059785832649245344787601820108477716142036228174305706008047], [1025130127158111391286114497628784310721816018761263921410307202662962155, 1046214728711544478045606234428908877874990102826941182072699504231466650], [702163857070541572271486478402213299769149650327011285713214313411220209, 964390034282681070472559926854266573294570007713258722688970731401446070], [545974440942587589793355643326618406027250242128080874758407596428510512, 761850948586179337743907452723161818650832861623511156392451479702387930], [1007560043455625957074409098454581056661266112007216804759867237444073697, 568301847329160573315400903846138202220379078132258736522271094761757556], [197721450177956824925214814076136877509231101020546157774722389794437045, 364710051200340597882807263315974397500183542853128578183482144214813413], [1040787630887710087842592894781064728294150449155777473810980176110345869, 209262206482771129037093161089704974346792207851381833556910011562779092], [817585719106595758704863067855703256513213471986817100166846551663829691, 537939259557282367210288215732191700156429531225372888960276076554578723]]\n","done":false}︡{"done":true}
︠e8188a1b-c1f4-468a-a099-d7aac9c67bc9︠
z=ECTimes(h,x,G)
z
d = [0]*len(c)
for counter in xrange(len(c)):
    d[counter] = ECAdd(c[counter],ECInverse(z,G),G)
d
ECUnembed(d,40)
︡7ad25eee-cad4-4f05-8544-df1981851545︡︡{"stdout":"[758836381415980725873880391356898132106948925841767677664020161182662238, 172885080124927030627775664387974993851680223072695154225919344264781309]\n","done":false}︡{"stdout":"[[7368168208605288208605287885288328448408128048565288048807888368488320, 121486482953176117528000809881749947571671902330970782925843120215624872], [8045845286928645288688608048605288368688328648208488328045288488448200, 903695374841292322370271174415204218208659055675797121100003842392500926], [8408648605845287368168045288488448208408648605287888568045288328208600, 452563156195183222674654981175834124068138520541249589841058897891287244], [8648048005288568208128168645287928048328448765287888408005288648168040, 354133454203841855971589741317387587553570850607256693528118869301689901], [8405288648168045288567888408008448365288408688367928048565287968168440, 566116327420959706426335734957260808053541488054698602617229920903468133], [8608048405287888408005288648168045288327888608645288448688648488688641, 190989246142380934866236751363212697612118794496744686603811386587621417], [5288208605288168448765288367888408845288488448208408648605287888568041, 425171620928964741474599056379821932316091568940530183178971060551595853], [5288408048048008048005840, 1036270550319895031743093327765070280126725147090778246179870977799274707]]\n","done":false}︡{"stdout":"'This is a longer example. It uses multiple points. The points are listed right below and then the random number chosen and the last output is how many points are needed.'\n","done":false}︡{"done":true}
︠02507fda-9411-4528-a15b-9f96961cec72︠

︡505626b6-0c82-4681-9242-987615d1ed7b︡︡{"done":true}
︠9ead035d-94e5-4e88-9268-4b9947281799︠


︡4a19e0b7-ccfb-441e-8773-049a5cb304f4︡︡{"done":true}









